﻿using System;
using System.Linq;
using System.Threading;
using System.Windows;

namespace EventComposition
{
    public partial class MainPage
    {
        #region Public Methods

        public MainPage()
        {
            InitializeComponent();

            Observable.Context = SynchronizationContexts.CurrentDispatcher;

            // this example shows how events can be combined with Rx
            IObservable<IObservable<int[]>> observerSelections = Observable.Merge(
                // multi valued
                GetDataStreamBySelectMany(),
                GetDataStreamByCombineLatest(),
                GetDataStreamByZip(),
                GetDataStreamByJoin(),
                GetDataStreamByJoin2(),
                GetDataStreamByForkJoin(),
                // single valued
                GetDataStreamByMerge(),
                GetDataStreamByAmb(),
                GetDataStreamByConcat(),
                GetDataStreamByUntil(),
                GetDataStreamByWaitUntil()
            );

            // clear ChoiceControl selections if another composition operator has been chosen
            IDisposable subscription = null;
            observerSelections.Subscribe(observer =>
            {
                if (subscription != null)
                    subscription.Dispose();

                choiceControl1.ClearAll();
                choiceControl2.ClearAll();
                choiceControl3.ClearAll();
                statusText.Text = String.Empty;

                subscription = observer.Subscribe(UpdateSelectedOptions);
            });
        }

        #endregion

        #region Private Helper Methods

        #region "Multi-valued" Compositions

        private IObservable<IObservable<int[]>> GetDataStreamBySelectMany()
        {
            //// using SelectMany directly
            //return Observable.FromEvent<RoutedEventArgs>(rbSelectMany, "Checked")
            //    .Select(_ => choiceControl1.OptionSelections.SelectMany(
            //        s1 => choiceControl2.OptionSelections.SelectMany(
            //        s2 => choiceControl3.OptionSelections.SelectMany(
            //        s3 => Observable.Return(new[] { s1, s2, s3 })))));

            // using "from" keyword => equal to use of SelectMany
            return Observable.FromEvent<RoutedEventArgs>(rbSelectMany, "Checked")
                .Select(_ => from s1 in choiceControl1.OptionSelections
                             from s2 in choiceControl2.OptionSelections
                             from s3 in choiceControl3.OptionSelections
                             select new[] { s1, s2, s3 });
        }

        private IObservable<IObservable<int[]>> GetDataStreamByCombineLatest()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbCombineLatest, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                             .CombineLatest(choiceControl2.OptionSelections,
                                            (sel1, sel2) => new[] { sel1, sel2 })
                             .CombineLatest(choiceControl3.OptionSelections,
                                            (selections, sel3) => new[] { selections[0], selections[1], sel3 }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByZip()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbZip, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                    .Zip(choiceControl2.OptionSelections, (sel1, sel2) => new[] { sel1, sel2 })
                    .Zip(choiceControl3.OptionSelections, (values, sel3) => new[] { values[0], values[1], sel3 }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByJoin()
        {
            // this is equal to Zip() of the 3 selection streams
            return Observable.FromEvent<RoutedEventArgs>(rbJoin, "Checked")
                .Select(_ => Observable.Join(
                                choiceControl1.OptionSelections
                                    .And(choiceControl2.OptionSelections)
                                    .And(choiceControl3.OptionSelections)
                                .Then((sel1, sel2, sel3) => new[] { sel1, sel2, sel3 })));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByJoin2()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbJoin2, "Checked")
                .Select(_ => Observable.Join(
                                choiceControl1.OptionSelections
                                    .And(choiceControl2.OptionSelections)
                                    .Then((sel1, sel2) => new[] { sel1, sel2 }),
                                choiceControl1.OptionSelections
                                    .And(choiceControl3.OptionSelections)
                                    .Then((sel1, sel3) => new[] { sel1, sel3 }),
                                choiceControl2.OptionSelections
                                    .And(choiceControl3.OptionSelections)
                                    .Then((sel2, sel3) => new[] { sel2, sel3 })));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByForkJoin()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbForkJoin, "Checked")
                .Select(_ => Observable.ForkJoin(
                                choiceControl1.OptionSelections,
                                choiceControl2.OptionSelections,
                                choiceControl3.OptionSelections));
        }

        #endregion

        #region "Single-valued" Compositions

        private IObservable<IObservable<int[]>> GetDataStreamByMerge()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbMerge, "Checked")
                .Select(_ => Observable.Merge(choiceControl1.OptionSelections,
                                              choiceControl2.OptionSelections,
                                              choiceControl3.OptionSelections)
                                       .Select(value => new[] { value }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByAmb()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbAmb, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                    .Amb(choiceControl2.OptionSelections)
                    .Amb(choiceControl3.OptionSelections)
                    .Select(value => new[] { value }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByConcat()
        {
            // attention: because we have infinite event streams, Concat will only forward
            // events from the first event stream, no matter which events are fired on stream 2 and 3
            return Observable.FromEvent<RoutedEventArgs>(rbConcat, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                    .Concat(choiceControl2.OptionSelections)
                    .Concat(choiceControl3.OptionSelections)
                    .Select(value => new[] { value }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByUntil()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbUntil, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                    .Until(choiceControl2.OptionSelections)
                    .Until(choiceControl3.OptionSelections)
                    .Select(value => new[] { value }));
        }

        private IObservable<IObservable<int[]>> GetDataStreamByWaitUntil()
        {
            return Observable.FromEvent<RoutedEventArgs>(rbWaitUntil, "Checked")
                .Select(_ => choiceControl1.OptionSelections
                    .WaitUntil(choiceControl2.OptionSelections)
                    .WaitUntil(choiceControl3.OptionSelections)
                    .Select(value => new[] { value }));
        }

        #endregion

        #region Other Helpers

        private void UpdateSelectedOptions(int[] values)
        {
            switch (values.Length)
            {
                case 1:
                    statusText.Text = String.Format("Values From Event: {0}", values[0]);
                    break;
                case 2:
                    statusText.Text = String.Format("Values From Event: {0}, {1}", values[0], values[1]);
                    break;
                case 3:
                    statusText.Text = String.Format("Values From Event: Option {0}, Option {1}, Option {2}",
                        values[0], values[1], values[2]);
                    break;
                default:
                    throw new InvalidOperationException("More than 3 values not supported (yet)");
            }
        }

        #endregion

        #endregion
    }
}
